
/*********************************************************************************
*										*
*    FILENAME      : NC_LOAD_3.C					     	*
*										*
*    PROJECT       : NI-MH Lader mit Konstantstrom via Analogkomparator        	*
*		     						               	*
*		                                                        	*
*		                                				*
*    KUNDE	   : 								*
*										*
*    BESCHREIBUNG  : 								*
*		     								*
*------------------------------------------------------------------------------	*
*										*
*    VERSION       : 0.3					   		*
*		          							*
*			  							*
*			  							*
*										*
*    BEARBEITER    : R.GION,  Hessingstr.2 86199 Augsburg			*
*		     Tel.: 0821-9988029 Fax: 0821-9988030 Mob.: 0177-6633804	*
*										*
*    DATUM         : 02.05.2008							*
*    MOD.          : 		   						*
*										*
********************************************************************************/

//	Chip type           : ATmega8
//	Program type        : Application
//	Clock frequency     : 8,000000 MHz (internal RC-oscillator !)
//	Memory model        : Small
//	External SRAM size  : 0
//	Data Stack size     : 256
//      Compiler            : CodeVision AVR Ver 1.24.6
//      Code memory used    : 10%
//      Programmer type     : AVR ISP, Pony-Prog
//
//*********************************************


#include <stdio.h>
#include <stdlib.h>
#include <mega8.h>

// while programming set corresponding fuse bit to activate internal RC-oscillator 8 MHz !
     
/********************************************************************************/
/*      I/O lines ATMEGA8	                                                */

// Port A

// Port B

#define SWITCH             PORTB.0	// Port B Pin 0 switcher control line

// Port C

// Port D

#define LED_1              PORTB.1	// Port B Pin 1 status led
#define LED_2              PORTB.2	// Port B Pin 2 status led
#define mode_1		    PIND.1
#define mode_2		    PIND.2
#define mode_3		    PIND.3

/********************************************************************************/

#define uchar	unsigned char
#define uint	unsigned int
#define sint	signed int

#define on		1
#define off		0

#define high            1
#define low             0


/********************************************************************************/
// declare variables

uchar   sys_time = 0;		// scheduler wrap around counter
uchar   batt_mess = 0;
uchar	trickle_counter = 0;
uint    elaps_counter = 0;
uint    adc_result = 0;
uint	batt_max = 0;
sint	batt_diff = 0;

/********************************************************************************/
// flags

bit current_end = 0;
bit charge_start = 0;
bit trickle = 0;
bit toggle = 0;

/********************************************************************************/
/* declare functions                                                            */

void sys_init(void);		        // Systemvariablen initialisieren
void ser_init(void);			// init serial port
void comparator_init(void);		// analog comparator init
void delay(unsigned long value);	//
uint get_adc(uchar chanel);		// get analog values
void current_on(void);			// turn on constant current
void current_off(void);			// turn off constant current
void check_start(void);			// check starting condition

/********************************************************************************/
void sys_init()
{
  DDRB = 0x07;		// Port B.0, B.1, B.2 as output
  DDRD = 0x00;		// Port D as input
  PORTD |= 0x0F;	// activate Port D.0, D.1, D.2, D.3 pullup
    
  SWITCH = high;        // set control line switcher to default
  LED_1 = low;          // led off
  
  ADCSRA=0x87;		// ADC enable - 125kHz
  
  TCCR0 |= 0x5;		// Timer 0 clock / 1024
  TCNT0  = 100;		// preset timer 0
  TIMSK |= 0x01;	// overflow int. enable
  
}
/********************************************************************************/
void ser_init()		// serial port transmit and receive 9600 baud
{                       
                                        
        UCSRA=0x00;
        UCSRB=0x98;
        UCSRC=0x86;
        UBRRH=0x00;
        UBRRL=0x67;
                                                          
}

/********************************************************************************/
void comparator_init()		// analog comparator init 
{                       
 ACSR |= 0x08;			// comparator interrupt enable
 
}
/********************************************************************************/
interrupt [ANA_COMP] void ana_comp_isr(void)   // analog comparator interrupt handling
{
  
  if ( ACSR & 0x20 ) 			// if current exeed limit turn off switcher pulse
  	SWITCH = low;
  else  SWITCH = high;
  	 
}
/*******************************************************************************/
interrupt [10] void timer0_overflow_int(void)        // overflow timer 0 ISR
{                                                      // every 15,625ms
  
  TCNT0 = 128;	// preset timer 0
  
  sys_time ++;   
    
  if ( sys_time & 1)			// <---- Interrupt 31,25ms
    {
       // place your code here if desired 
    }
  
  else
/*********************************************************************************/   
  if ( sys_time & 2)	      		// <---- Interrupt 62,5ms
    {
       // place your code here if desired
    }
    
  else
  
  /*********************************************************************************/ 
  
  if ( sys_time & 4)	    		// <---- Interrupt 125ms
   {
     if ( ! charge_start )
      	LED_1 = off;
      
   }  
   
  else
  
  /*********************************************************************************/
      
  if ( sys_time & 8)	    		// <---- Interrupt 250ms
    {
      if ( ! charge_start )
      	LED_1 = on;
          
    } 
  
  else
      
  /*********************************************************************************/
  
  if ( sys_time & 16)  			// <---- Interrupt 500ms
    {
       if ( ! current_end & ! toggle & charge_start )
       		{
       			LED_1 = on;
       			toggle = 1;
       		}
       	else
       	
       	if ( ! current_end & toggle & charge_start )
       		{
       			LED_1 = off;
       			toggle = 0;
       		}	
        	                   	  
    } 
  
  else
  
  /*********************************************************************************/
  
  if ( sys_time & 32)			// <---- Interrupt 1000ms
    {
     batt_mess ++;
     
     if ( trickle )
     	trickle_counter ++;
    } 
     
}

/********************************************************************************/
void delay (unsigned long time)	// Delay = (time * 1ms)  @8MHz
{
 time *= 180;
 
 while (time > 0)
      time --;
}                                                                                
/********************************************************************************/
uint get_adc(uchar chanel)
{
  ADMUX = chanel;	// select chanel
  ADCSRA |= 0x40;	// start conversation
  
  while ( ADCSRA & 0x40);// wait for conversion finished
  adc_result = ADCW;
  return(adc_result);  
}
/********************************************************************************/
void current_off()	// turns off constant current stage
{
  ACSR &= 0x00;		// disable analog comparator interrupt
  ACSR |=0x10;		// clear pending interrupt flag
  SWITCH = low;		// disable constant current switch

}
/********************************************************************************/
void current_on()       // turns on constant current stage
{
  ACSR |= 0x08;		// enable analog comparator
  SWITCH = high;        // enable constant current switch

}
/********************************************************************************/
void check_start()
{
  uint v_batt;
  
  delay(1000);		// wait 1s
  v_batt = get_adc(0);	// get battery voltage		
  
  if ( v_batt <=50 | v_batt >= 1000 ) 	// if too low or too high then there is a
  	charge_start = 0;		// short circuit or no battery is connected
  else charge_start = 1;
  
}
/********************************************************************************/
/*										*/
/*			     m a i n    p r o g r a m				*/
/*										*/
/********************************************************************************/

main()

{
  uint v_batt;
  
  sys_init();                           // init system, timer
  //ser_init();				// used for debug purpose only !
  comparator_init();			// analog comparator init
  SREG |= 0x80;				// global enable interrupt
  current_off();			// turn off constant current			
      
  while ( ! charge_start )		// check if battery is connected
  	check_start();
  
  
  current_on();				// if battery is ok then start charging
  
  
  for(;;)                         	// loop forever !
  {                             		
      // main charging procedur
      
      if ( batt_mess >= 15 & !current_end )	// check every 15s
      	{
      		batt_mess = 0;			// clear counter
      		
      		current_off();			// turn off charging current
      		delay(1000);			// wait 500ms
      		v_batt = get_adc(0);		// read ADC
      		      		
      		if ( v_batt >= batt_max )	// check for new peak voltage
      			batt_max = v_batt;  	// store new peak value
      		
      		batt_diff = batt_max - v_batt;  // build difference
      		      		
      		if ( batt_diff >= 10 )		// if voltage drops more then 10 units
      		{
      			current_end = 1;        // then turn off charging any longer
      			LED_1 = on;
      			current_off();		// turn off current
      			if ( mode_1 )		// check if trickle is used
      				trickle = 1; 	// enable trickle charge
      				
      		}	
     		
     		     		
     		if ( ! current_end )
     		     current_on();		// turn on charging current
     		
     	}	
     	  
        // if selected by input pin then start trickle charge
        
        if ( trickle_counter >= 5 )		// happens every 5s
        	{
        		
        		trickle_counter = 0;	// clear counter
        		current_on();		// turn on current pulse 1s
        		LED_1 = on;		// show this with the LED_1
        		delay(1000);		// wait simply 1s
        		LED_1 = off;		// turn off LED_1 and current
        		current_off();
        		elaps_counter ++;	// increment repeat counter
        		
        		if (elaps_counter >= 480 ) // if repeat rate reached then quit
        			{
        				trickle = 0;	// clear trickle mode
        				LED_1 = on;	// steady light for full charge
        			}
    	
       		}
    
    } 
  
}

/********************************************************************************/

/* end of source file */